from typing import List
from copy import deepcopy

import torch
import torch.nn as nn
import numpy as np


class GeneratedText:
    def __init__(
        self,
        current_tokens: List[int],
        initial_logprob: float = 0,
        vocab_size: int = 60000,
        device=torch.device("cpu"),
    ):
        self.tokens = current_tokens
        self.generated_tokens = []
        self.logprob = initial_logprob
        self.vocab_size = vocab_size
        self.repetition_penalizer = torch.ones((vocab_size,)).to(device)

    def add_new_token(self, token, logprob=0, repetition_penalty: float = 1.0):
        # if token != self.tokens[-1]:
        self.tokens.append(token)
        self.generated_tokens.append(token)
        self.logprob += logprob
        self.repetition_penalizer[token] += repetition_penalty

    def __len__(self):
        return len(self.tokens)

    def get_tokens_for_decoder(self, decoder_max_len):
        if len(self.tokens) > decoder_max_len:
            # half = decoder_max_len // 2
            # return self.tokens[:half] + self.tokens[-half:]
            return self.tokens[-decoder_max_len:]
        else:
            return self.tokens

    def get_repetition_penalizer(self):
        return self.repetition_penalizer


class LMPipeline:
    def __init__(
        self,
        tokenizer,
        sos_token: int = 0,
        eos_token: int = 1,
        vocab_size: int = 60000,
    ):
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.tokenizer = tokenizer
        self.vocab_size = vocab_size

    @torch.no_grad()
    def new_decoder_sample(
        self,
        model: torch.nn.Module,
        tokenized_input,
        do_sample: bool = False,
        temperature: float = 1.0,
        top_k: int = None,
        num_beams: int = 1,
        repetition_penalizer: torch.Tensor = torch.ones(
            1,
        ),
        device=torch.device("cpu"),
        cross_attention_tokens: torch.Tensor = None,
    ):
        decoder_input = (
            torch.Tensor(tokenized_input).type(torch.int).unsqueeze(0).to(device)
        )
        if cross_attention_tokens is not None:
            logits = model.get_logits_next_token(decoder_input, cross_attention_tokens)
        else:
            logits = model.get_logits_next_token(decoder_input)
        last = logits.reshape((-1,))
        last = last / (temperature * repetition_penalizer)
        if num_beams > 1:
            v, beam_top_indices = torch.topk(last, k=top_k, sorted=False, largest=True)
            last[last < v.min()] = -float("Inf")
        elif top_k is not None:
            v, _ = torch.topk(last, k=top_k, sorted=False, largest=True)
            last[last < v.min()] = -float("Inf")

        probas = nn.functional.softmax(last, dim=0)

        chosen = []
        logprob = []
        for i in range(num_beams):
            if do_sample:
                token_chosed = torch.multinomial(probas, 1).detach().cpu().item()
            else:
                token_chosed = torch.argmax(probas).detach().cpu().item()

            logprob_token = 1 - torch.log(probas[token_chosed]).detach().cpu().item()
            chosen.append(token_chosed)
            logprob.append(logprob_token)

        return chosen, logprob

    def decoder_standard_generation(
        self,
        model: torch.nn.Module,
        input_text: str,
        max_tokens: int,
        decoder_max_len: int = 512,
        do_sample: bool = False,
        temperature: float = 1.0,
        top_k: int = None,
        num_breams: int = 1,
        repetition_penalty: float = 1,
        device=torch.device("cpu"),
        cross_attention_tokens: torch.Tensor = None,
        insert_bos_decoder: bool = True,
    ):
        tokenized = self.tokenizer.tokenize_text(
            input_text, padding=False, truncation=False
        )
        if insert_bos_decoder:
            tokenized = [self.sos_token] + tokenized

        generated_text = GeneratedText(
            current_tokens=tokenized, vocab_size=self.vocab_size, device=device
        )
        generated_texts = [generated_text]
        for _ in range(max_tokens):
            new_generated_texts = []
            for current_text in generated_texts:
                # se ja tiver chegado no final esse cara nao adiciona mais nada
                if current_text.tokens[-1] == self.eos_token:
                    new_generated_texts.append(deepcopy(current_text))
                    continue

                chosen, logprob = self.new_decoder_sample(
                    model=model,
                    tokenized_input=current_text.get_tokens_for_decoder(
                        decoder_max_len=decoder_max_len
                    ),
                    do_sample=do_sample,
                    temperature=temperature,
                    top_k=top_k,
                    num_beams=num_breams,
                    repetition_penalizer=current_text.get_repetition_penalizer(),
                    device=device,
                    cross_attention_tokens=cross_attention_tokens,
                )

            for new_token, token_logprob in zip(chosen, logprob):
                if new_token != self.eos_token:
                    new_text = deepcopy(current_text)
                    new_text.add_new_token(
                        new_token, token_logprob, repetition_penalty=repetition_penalty
                    )
                    new_generated_texts.append(new_text)
                else:
                    new_generated_texts.append(current_text)

            generated_texts = self.filter_beam_generated(
                new_generated_texts, num_beams=num_breams**3, normalize_prob=True
            )

        generated_text = self.filter_beam_generated(
            generated_texts, num_beams=1, normalize_prob=True
        )[0]
        return self.tokenizer.untokenize_tokens(generated_text.tokens)

    def filter_beam_generated(
        self, texts: List[GeneratedText], num_beams: int, normalize_prob: bool = True
    ):
        if len(texts) <= num_beams:
            return texts

        probas = [t.logprob for t in texts]
        if normalize_prob:
            probas = [p / (len(t.tokens) ** 0.7) for p, t in zip(probas, texts)]

        filtered = []
        for i in range(num_beams):
            biggest = np.argmax(probas)
            filtered.append(texts[biggest])
            probas.pop(biggest)
            texts.pop(biggest)

        return filtered

    @torch.no_grad()
    def decoder_nucleus_generation(
        self,
        model: torch.nn.Module,
        input_text: str,
        max_tokens: int,
        decoder_max_len: int = 512,
        p: float = 0.95,
        temperature: float = 1.0,
        repetition_penalty: float = 1.0,
        device=torch.device("cpu"),
        cross_attention_tokens: torch.Tensor = None,
    ):
        tokenized = [self.sos_token] + self.tokenizer.tokenize_text(
            input_text, padding=False, truncation=False
        )
        generated_text = GeneratedText(
            current_tokens=tokenized, vocab_size=self.vocab_size, device=device
        )

        for i in range(max_tokens):
            decoder_input = (
                torch.Tensor(
                    generated_text.get_tokens_for_decoder(
                        decoder_max_len=decoder_max_len
                    )
                )
                .type(torch.int)
                .unsqueeze(0)
                .to(device)
            )
            if cross_attention_tokens is not None:
                logits = model.get_logits_next_token(
                    decoder_input, cross_attention_tokens
                )
            else:
                logits = model.get_logits_next_token(decoder_input)
            last = logits.reshape((-1,))
            probas = nn.functional.softmax(
                last / (temperature * generated_text.get_repetition_penalizer()), dim=0
            )
            # v, top_indices = torch.topk(probas, k=1000, sorted=True, largest=True)
            sorted_probas, sorted_indices = torch.sort(probas, descending=True, dim=-1)
            cumulative_sum = torch.cumsum(sorted_probas, dim=-1)
            # pode acontecer da primeira probabilidade ja ser maior que p
            if cumulative_sum[0] > p:
                p_aplicado = 1
            else:
                p_aplicado = p

            selected = cumulative_sum <= p_aplicado

            nucleus_probas = sorted_probas.masked_fill(selected == 0, 0)
            nucleus_probas = nucleus_probas / torch.sum(nucleus_probas)
            # nucleus_probas = torch.softmax(nucleus_probas, dim=-1)

            # print(self.tokenizer.untokenize_tokens(nucleus))
            token_chosed = torch.multinomial(nucleus_probas, 1).detach().cpu().item()
            new_token = sorted_indices[token_chosed].detach().cpu().item()
            generated_text.add_new_token(
                new_token, repetition_penalty=repetition_penalty
            )

            if new_token == self.eos_token:
                break

        return self.tokenizer.untokenize_tokens(generated_text.tokens)



# implementacao do paper https://arxiv.org/pdf/2205.12035

import torch
import torch.nn as nn

from models.base_encoder.model import EncoderMLM
from transformer.transformer_blocks import (
    PositionalEncoding,
    MultiHeadCrossAttention,
    FeedFowardBlock,
)


class RetroMaeDecoder(nn.Module):
    """
    Por conta da forma como eh feita a atencao nele (nao eh causal), vamos
    usar um bloco de encoder
    """

    def __init__(self, decoder_args):
        super().__init__()

        self.masking_ratio = decoder_args["masking_ratio"]

        self.pos_encoding = PositionalEncoding(
            model_dimension=decoder_args["embed_dim"], dropout_probability=0.0
        )

        self.embeddings = nn.Embedding(
            num_embeddings=decoder_args["vocab_size"],
            embedding_dim=decoder_args["embed_dim"],
        )

        self.attention = MultiHeadCrossAttention(
            num_heads=decoder_args["num_heads"],
            embed_dim=decoder_args["embed_dim"],
            dropout=decoder_args["dropout"],
        )

        self.ff = FeedFowardBlock(
            embed_dim=decoder_args["embed_dim"],
            hidden_size=decoder_args["hidden_size"],
            dropout=decoder_args["dropout"],
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=decoder_args["embed_dim"])
        self.norm_2 = nn.LayerNorm(normalized_shape=decoder_args["embed_dim"])

        self.lm_proj = nn.Linear(decoder_args["embed_dim"], decoder_args["vocab_size"])

    def forward(self, enc_emb: torch.Tensor, decoder_input):
        """
        enc_emb - mean embedding from encoder [B, embed_dim]
        decoder_input - raw tokens [B, input_len]
        """
        batch_size, input_len = decoder_input.shape
        # muda o shape para [B, 1, embed_dim] e depois [B, input_len, embed_dim]
        h1 = enc_emb.unsqueeze(1).repeat((1, input_len, 1))
        # adiciona o positional encoding
        h1 = self.pos_encoding(h1)

        h2 = self.embeddings(decoder_input)
        # nao usa a primeira posicao, substituimos pelo embedding do encoder
        h2[:, 0, :] = enc_emb

        # por ultimo a mascara da atencao
        # so nao presta atencao em si mesmo
        att_mask = torch.rand((batch_size, input_len, input_len))
        # precisa ter certeza que nenhum deles vai prestar atencao em si mesmo
        att_mask -= torch.eye(input_len)
        # precisa garantir que todos menos a primeira linha prestem atencao no token 0 (encoder embedding)
        att_mask[:, 1:, 0] = 1.0
        att_mask = torch.masked_fill(
            att_mask, att_mask <= self.masking_ratio, float("-inf")
        )
        att_mask = torch.masked_fill(att_mask, att_mask > float("-inf"), 0)
        # shape [input_len, input_len] -> [1, 1, input_len, input_len]
        att_mask = att_mask.reshape((batch_size, 1, input_len, input_len))
        att_mask = att_mask.to(decoder_input.device)

        # agora roda
        x = self.norm_1(h1 + self.attention(h1, h2, att_mask))
        x = self.norm_2(x + self.ff(x))
        x = self.lm_proj(x)
        return x  # lm logits


class RetroMaeModel(nn.Module):
    def __init__(self, encoder_args, decoder_args):
        super().__init__()

        self.encoder = EncoderMLM(**encoder_args)
        self.decoder = RetroMaeDecoder(decoder_args)

    def forward(self, x_masked, x_mask, x_unmasked):
        encoder_logits, embeddings = self.encoder(
            x_masked, x_mask, return_hidden_states=True
        )
        sentence_embedding = embeddings[:, 0, :]

        decoder_logits = self.decoder(sentence_embedding, x_unmasked)

        return encoder_logits, decoder_logits

    def embed(self, x):
        return self.encoder.embed(x)



from dataclasses import dataclass

import torch
import torch.nn as nn

from charylumodels.transformer.attention_blocks import RotaryMultiHeadFlashAttention
from charylumodels.transformer.transformer_blocks import FeedFowardBlock
from charylumodels.transformer.rotary import precompute_freqs_cis


@dataclass
class ModelArgs:
    vocab_size: int
    num_layers: int
    embed_dim: int
    hidden_size: int
    num_heads: int
    dropout: float = 0.1
    window_size: int = -1
    rotary_theta: int = 10_000

    max_seq_len: int = 2048


class EncoderBlock(nn.Module):
    def __init__(self, params: ModelArgs):
        super().__init__()

        self.attention = RotaryMultiHeadFlashAttention(
            num_heads=params.num_heads,
            embed_dim=params.embed_dim,
            dropout=params.dropout,
            causal=False,
            window_size=params.window_size,
        )
        self.feedforward = FeedFowardBlock(
            embed_dim=params.embed_dim,
            hidden_size=params.hidden_size,
            dropout=params.dropout,
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=params.embed_dim)
        self.norm_2 = nn.LayerNorm(normalized_shape=params.embed_dim)
        self.drop_skip_1 = nn.Dropout(params.dropout)
        self.drop_skip_2 = nn.Dropout(params.dropout)

    def forward(self, x, rotary_freqs):
        x = self.drop_skip_1(x) + self.attention(self.norm_1(x), rotary_freqs)
        x = self.drop_skip_2(x) + self.feedforward(self.norm_2(x))
        return x


class Encoder(nn.Module):
    def __init__(self, params: ModelArgs):
        super().__init__()

        self.rotary_freqs = precompute_freqs_cis(
            dim=params.embed_dim // params.num_heads,
            end=params.max_seq_len,
            theta=params.rotary_theta,
        )
        self.embedding = nn.Embedding(
            num_embeddings=params.vocab_size, embedding_dim=params.embed_dim
        )

        self.layers = nn.ModuleList()
        for _ in range(params.num_layers):
            self.layers.append(EncoderBlock(params))

        self.output_norm = nn.LayerNorm(normalized_shape=params.embed_dim)

    def forward(self, x):
        B, seq_len = x.shape
        rotary_freqs = self.rotary_freqs[:seq_len].to(x.device)

        x = self.embedding(x)

        for layer in self.layers:
            x = layer(x, rotary_freqs)

        # ja retorna uma saida normalizada
        # nenhum wrapper precisa normalizar a saida
        return self.output_norm(x)


class EncoderMLM(nn.Module):
    def __init__(self, params: ModelArgs):
        super().__init__()

        self.embed_dim = params.embed_dim

        self.encoder = Encoder(params=params)

        self.mlm_head = nn.Linear(params.embed_dim, params.vocab_size)

    def forward(self, x, mask, return_hidden_states: bool = False):
        # first get the masked_ids to use later
        # flattens the masked id so its easier to deal with
        masked_ids = torch.flatten(mask.reshape((-1,)).nonzero())
        # get all hidden states (they are already normalized)
        last_hidden_states = self.encoder(x)
        # flatten everything so we can compute everything at once
        all_hidden_states = last_hidden_states.reshape((-1, self.embed_dim))
        # get only the masked hidden states
        masked_hidden_states = all_hidden_states[masked_ids, :]
        # predicts only the masked tokens
        logits = self.mlm_head(masked_hidden_states)

        if return_hidden_states:
            return logits, last_hidden_states

        return logits

    def embed(self, x):
        return self.encoder(x)



import torch
import torch.nn as nn
import numpy as np

from charylumodels.transformer.activations import NewGELU
from charylumodels.transformer.attention import qkv_attention


class PositionalEncoding(nn.Module):
    def __init__(
        self, model_dimension, dropout_probability, expected_max_sequence_length=5000
    ):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout_probability)

        position_id = torch.arange(0, expected_max_sequence_length).unsqueeze(1)
        frequencies = torch.pow(
            10000.0,
            -torch.arange(0, model_dimension, 2, dtype=torch.float) / model_dimension,
        )

        positional_encodings_table = torch.zeros(
            expected_max_sequence_length, model_dimension
        )
        positional_encodings_table[:, 0::2] = torch.sin(
            position_id * frequencies
        )  # sine on even positions
        positional_encodings_table[:, 1::2] = torch.cos(
            position_id * frequencies
        )  # cosine on odd positions

        self.register_buffer("positional_encodings_table", positional_encodings_table)

    def forward(self, embeddings_batch):
        assert (
            embeddings_batch.ndim == 3
            and embeddings_batch.shape[-1] == self.positional_encodings_table.shape[1]
        ), f"Expected (batch size, max token sequence length, model dimension) got {embeddings_batch.shape}"

        positional_encodings = self.positional_encodings_table[
            : embeddings_batch.shape[1]
        ]

        return self.dropout(embeddings_batch + positional_encodings)


class MultiHeadAttention(nn.Module):
    def __init__(self, num_heads: int = 8, embed_dim: int = 512, dropout=0.1):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        # precisa virar [batch * heads, len, head_dim]
        x = x.contiguous().view((B, L, self.num_heads, self.head_dim)).transpose(1, 2)
        # virou [batch, heads, len, head_dim]
        # x = x.contiguous().view((B * self.num_heads, L, self.head_dim))
        return x

    def reshape_from_attention(self, x):
        B, H, L, HD = x.shape
        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x = x.transpose(1, 2)
        # virou [batch, len, heads, head_dim]
        x = x.contiguous().view((B, L, self.embed_dim))
        # virou [batch, len, embed_dim]
        return x

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None):
        q = self.reshape_for_attention(self.proj_q(x))
        k = self.reshape_for_attention(self.proj_k(x))
        v = self.reshape_for_attention(self.proj_v(x))

        x_att = qkv_attention(q, k, v, mask)
        # dropout
        x_att = self.dropout_attention(x_att)

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att = self.reshape_from_attention(x_att)

        # projecao final
        x_att = self.out_projection(x_att)
        x_att = self.dropout_projection(x_att)
        return x_att


class MultiHeadCrossAttention(nn.Module):
    def __init__(self, num_heads: int = 8, embed_dim: int = 512, dropout=0.1):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        # precisa virar [batch * heads, len, head_dim]
        x = x.contiguous().view((B, L, self.num_heads, self.head_dim)).transpose(1, 2)
        # virou [batch, heads, len, head_dim]
        # x = x.contiguous().view((B * self.num_heads, L, self.head_dim))
        return x

    def reshape_from_attention(self, x):
        B, H, L, HD = x.shape
        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x = x.transpose(1, 2)
        # virou [batch, len, heads, head_dim]
        x = x.contiguous().view((B, L, self.embed_dim))
        # virou [batch, len, embed_dim]
        return x

    def forward(self, x: torch.Tensor, y: torch.Tensor, mask: torch.Tensor = None):
        q = self.reshape_for_attention(self.proj_q(x))
        k = self.reshape_for_attention(self.proj_k(y))
        v = self.reshape_for_attention(self.proj_v(y))

        x_att = qkv_attention(q, k, v, mask)
        x_att = self.dropout_attention(x_att)

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att = self.reshape_from_attention(x_att)

        # projecao final
        x_att = self.out_projection(x_att)
        x_att = self.dropout_projection(x_att)
        return x_att


class FeedFowardBlock(nn.Module):
    def __init__(self, embed_dim, hidden_size, dropout: float = 0.1):
        super().__init__()

        self.ff_1 = nn.Linear(embed_dim, hidden_size)
        self.dropout = nn.Dropout(dropout)
        self.ff_2 = nn.Linear(hidden_size, embed_dim)
        self.activation = NewGELU()

    def forward(self, x):
        x_1 = self.ff_1(x)
        x_act = self.activation(x_1)
        x_drop = self.dropout(x_act)
        x_2 = self.ff_2(x_drop)
        return x_2


class DecoderBlock(nn.Module):
    def __init__(self, embed_dim, num_heads, hidden_size, dropout: float = 0.1):
        super().__init__()

        self.attention = MultiHeadAttention(num_heads=num_heads, embed_dim=embed_dim)
        self.feedforward = FeedFowardBlock(
            embed_dim=embed_dim, hidden_size=hidden_size, dropout=dropout
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=embed_dim)
        self.norm_2 = nn.LayerNorm(normalized_shape=embed_dim)
        self.drop_skip_1 = nn.Dropout(dropout)
        self.drop_skip_2 = nn.Dropout(dropout)

    def forward(self, x, mask):
        x = self.drop_skip_1(x) + self.attention(self.norm_1(x), mask)
        x = self.drop_skip_2(x) + self.feedforward(self.norm_2(x))
        return x


class EncoderBlock(nn.Module):
    def __init__(self, embed_dim, num_heads, hidden_size, dropout: float = 0.1):
        super().__init__()

        self.attention = MultiHeadAttention(
            num_heads=num_heads, embed_dim=embed_dim, dropout=dropout
        )
        self.feedforward = FeedFowardBlock(
            embed_dim=embed_dim, hidden_size=hidden_size, dropout=dropout
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=embed_dim)
        self.norm_2 = nn.LayerNorm(normalized_shape=embed_dim)
        self.drop_skip_1 = nn.Dropout(dropout)
        self.drop_skip_2 = nn.Dropout(dropout)

    def forward(self, x, mask=None):
        x = self.norm_1(self.drop_skip_1(x) + self.attention(x, mask))
        x = self.norm_2(self.drop_skip_2(x) + self.feedforward(x))
        return x



from typing import List
from pathlib import Path


import torch
import torch.nn as nn
import numpy as np
from tqdm import tqdm

from tokenizers.tokenizer import MyTokenizer


@torch.no_grad()
class MLMPipeline:
    def __init__(
        self,
        mask_token_id: int,
        tokenizer,
        model: torch.nn.Module,
        sos_token: int = 0,
        eos_token: int = 1,
    ):
        self.mask_token_id = mask_token_id
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.tokenizer = tokenizer

        self.model = model
        self.model.eval()

    def predict_mask(
        self,
        sentence: str,
        device,
        top_k: int = 5,
        max_length=512,
        padding="max_length",
        truncation=True,
        pad_input: bool = False,
        pad_token_id: int = 0,
        pad_multiple_of: int = 31,
    ):
        tokenized = self.tokenizer.tokenize_text(
            sentence, padding=padding, max_length=max_length, truncation=truncation
        )
        if pad_input:
            desired_size = (len(tokenized) // pad_multiple_of + 1) * pad_multiple_of
            diff = desired_size - len(tokenized)
            tokenized = [pad_token_id for _ in range(diff)] + tokenized
        # print(self.tokenizer.untokenize_tokens(tokenized))
        tokenized_t = torch.Tensor(tokenized).type(torch.int).to(device)
        masked = tokenized_t == self.mask_token_id
        # masked_indexes = torch.flatten(masked.nonzero())
        # print(tokenized[masked_indexes[0]])
        output = self.model(tokenized_t.unsqueeze(dim=0), masked).squeeze()
        if len(output.shape) < 2:
            output = torch.unsqueeze(output, dim=0)
        # pega o top 5 preditado
        ordered = torch.argsort(torch.softmax(output, dim=1), descending=True)
        if len(ordered.shape) > 1:
            top = [ordered[i][:top_k].tolist() for i in range(ordered.shape[0])]
        else:
            top = [torch.argsort(output, descending=True)[:top_k].tolist()]

        respostas = []
        tokens = []
        probas = []
        for i in range(len(top)):
            tokens.append([])
            probas.append([])
            for k in range(top_k):
                proba = torch.softmax(output, dim=1)[i, top[i][k]]
                token_preditado = self.tokenizer.untokenize_tokens([top[i][k]])
                respostas.append(f"token {i} - {k}: {token_preditado} - {proba:0.5f}")
                tokens[i].append(token_preditado)
                probas[i].append(proba.detach().cpu())
        return respostas, tokens, probas

    def print_pretty(self, input_text, respostas, tokens, probas, top_ans: int = 5):
        probas = np.array(probas)
        tokens = np.array(tokens)
        quantidade_masks = len(tokens)

        for top in range(top_ans):
            nova_frase = input_text
            for m_id in range(quantidade_masks):
                nova_frase = nova_frase.replace("<mask>", tokens[m_id, top], 1)

            print(f"PROBAS: {probas[:, top]}")
            print(nova_frase)

        for r in respostas:
            print(r)



import torch
import torch.nn as nn
import numpy as np

from flash_attn import flash_attn_func

from charylumodels.transformer.rotary import apply_rotary_emb, apply_cross_rotary_emb
from charylumodels.transformer.attention import qkv_attention, make_causal_mask
from charylumodels.transformer.transformer_blocks import FeedFowardBlock
from charylumodels.transformer.norm import RMSNorm


class RotaryMultiHeadAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        # precisa virar [batch * heads, len, head_dim]
        x = x.contiguous().view((B, L, self.num_heads, self.head_dim)).transpose(1, 2)
        # virou [batch, heads, len, head_dim]
        # x = x.contiguous().view((B * self.num_heads, L, self.head_dim))
        return x

    def reshape_from_attention(self, x):
        B, H, L, HD = x.shape
        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x = x.transpose(1, 2)
        # virou [batch, len, heads, head_dim]
        x = x.contiguous().view((B, L, self.embed_dim))
        # virou [batch, len, embed_dim]
        return x

    def forward(
        self, x: torch.Tensor, rotary_freqs: torch.Tensor, mask: torch.Tensor = None
    ):
        q = self.proj_q(x)
        k = self.proj_k(x)
        v = self.reshape_for_attention(self.proj_v(x))

        # rotary embeddings
        q, k = apply_rotary_emb(q, k, rotary_freqs)

        q = self.reshape_for_attention(q)
        k = self.reshape_for_attention(k)

        x_att = qkv_attention(q, k, v, mask)
        x_att = self.dropout_attention(x_att)

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att = self.reshape_from_attention(x_att)

        # projecao final
        x_att = self.out_projection(x_att)
        x_att = self.dropout_projection(x_att)
        return x_att


class RotaryMultiHeadFlashAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
        causal: bool = False,
        window_size: int = -1,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.causal = causal
        self.window_size = window_size  # new feature from flash attention

        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        x_view = x.contiguous().view((B, L, self.num_heads, self.head_dim))
        # virou [batch, len, heads, head_dim]
        return x_view

    def reshape_from_attention(self, x):
        B, L, H, HD = x.shape
        # virou [batch, len, heads, head_dim]
        x_view = x.contiguous().view(B, L, self.embed_dim)
        # virou [batch, len, embed_dim]
        return x_view

    def forward(
        self,
        x: torch.Tensor,
        rotary_freqs: torch.Tensor,
    ):
        q = self.reshape_for_attention(self.proj_q(x.clone()))
        k = self.reshape_for_attention(self.proj_k(x.clone()))
        v = self.reshape_for_attention(self.proj_v(x.clone()))

        # rotary embeddings
        q, k = apply_rotary_emb(q, k, rotary_freqs)

        if next(self.parameters()).is_cuda:
            x_att = flash_attn_func(
                q=q.half(),  # flash requer float16
                k=k.half(),
                v=v.half(),
                dropout_p=self.dropout if self.training else 0.0,
                softmax_scale=1 / np.sqrt(self.head_dim),
                causal=self.causal,
                return_attn_probs=False,
                window_size=(self.window_size, self.window_size),
            )
        else:
            x_att = qkv_attention(
                q.transpose(1, 2),
                k.transpose(1, 2),
                v.transpose(1, 2),
                mask=make_causal_mask(q) if self.causal else None,
            ).transpose(1, 2)
        # para inferencia
        if not self.training or q.dtype != x_att.dtype:
            # se o modelo esta em float32 precisa voltar para float32
            for parametro in self.out_projection.parameters():
                x_att_ajustado = x_att.type(parametro.dtype)
                break
        else:
            x_att_ajustado = x_att

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att_reshaped = self.reshape_from_attention(x_att_ajustado)

        # projecao final
        x_att_projected = self.out_projection(self.dropout_projection(x_att_reshaped))

        return x_att_projected


class RotaryMHFlashCrossAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
        causal: bool = False,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.causal = causal

        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        x_view = x.contiguous().view((B, L, self.num_heads, self.head_dim))
        # virou [batch, len, heads, head_dim]
        return x_view

    def reshape_from_attention(self, x):
        B, L, H, HD = x.shape
        # virou [batch, len, heads, head_dim]
        x_view = x.contiguous().view(B, L, self.embed_dim)
        # virou [batch, len, embed_dim]
        return x_view

    def forward(
        self,
        x: torch.Tensor,
        rotary_freqs: torch.Tensor,
    ):
        q = self.reshape_for_attention(self.proj_q(x.clone()))
        k = self.reshape_for_attention(self.proj_k(x.clone()))
        v = self.reshape_for_attention(self.proj_v(x.clone()))

        # rotary embeddings
        q, k = apply_cross_rotary_emb(q, k, rotary_freqs)

        x_att = flash_attn_func(
            q=q.half(),  # flash requer float16
            k=k.half(),
            v=v.half(),
            dropout_p=self.dropout if self.training else 0.0,
            softmax_scale=1 / np.sqrt(self.head_dim),
            causal=self.causal,
            return_attn_probs=False,
            window_size=(self.window_size, self.window_size),
        )
        # para inferencia
        if not self.training or q.dtype != x_att.dtype:
            # se o modelo esta em float32 precisa voltar para float32
            for parametro in self.out_projection.parameters():
                x_att_ajustado = x_att.type(parametro.dtype)
                break
        else:
            x_att_ajustado = x_att

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att_reshaped = self.reshape_from_attention(x_att_ajustado)

        # projecao final
        x_att_projected = self.out_projection(self.dropout_projection(x_att_reshaped))

        return x_att_projected


# blocos sem relative positional encoding mas com flash attention
class MultiHeadFlashAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
        causal: bool = False,
        window_size: int = -1,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.causal = causal
        self.window_size = window_size

        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        x_view = x.contiguous().view((B, L, self.num_heads, self.head_dim))
        # virou [batch, len, heads, head_dim]
        return x_view

    def reshape_from_attention(self, x):
        B, L, H, HD = x.shape
        # virou [batch, len, heads, head_dim]
        x_view = x.contiguous().view(B, L, self.embed_dim)
        # virou [batch, len, embed_dim]
        return x_view

    def forward(
        self,
        x: torch.Tensor,
    ):
        q = self.reshape_for_attention(self.proj_q(x.clone()))
        k = self.reshape_for_attention(self.proj_k(x.clone()))
        v = self.reshape_for_attention(self.proj_v(x.clone()))

        x_att = flash_attn_func(
            q=q.half(),  # flash requer float16
            k=k.half(),
            v=v.half(),
            dropout_p=self.dropout if self.training else 0.0,
            softmax_scale=1 / np.sqrt(self.head_dim),
            causal=self.causal,
            return_attn_probs=False,
            window_size=(self.window_size, self.window_size),
        )
        # para inferencia
        if not self.training or q.dtype != x_att.dtype:
            # se o modelo esta em float32 precisa voltar para float32
            for parametro in self.out_projection.parameters():
                x_att_ajustado = x_att.type(parametro.dtype)
                break
        else:
            x_att_ajustado = x_att

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att_reshaped = self.reshape_from_attention(x_att_ajustado)

        # projecao final
        x_att_projected = self.out_projection(self.dropout_projection(x_att_reshaped))

        return x_att_projected


class MultiHeadFlashCrossAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
        causal: bool = False,
        window_size: int = -1,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.causal = causal
        self.window_size = window_size

        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        x_view = x.contiguous().view((B, L, self.num_heads, self.head_dim))
        # virou [batch, len, heads, head_dim]
        return x_view

    def reshape_from_attention(self, x):
        B, L, H, HD = x.shape
        # virou [batch, len, heads, head_dim]
        x_view = x.contiguous().view(B, L, self.embed_dim)
        # virou [batch, len, embed_dim]
        return x_view

    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ):
        q = self.reshape_for_attention(self.proj_q(x.clone()))
        k = self.reshape_for_attention(self.proj_k(y.clone()))
        v = self.reshape_for_attention(self.proj_v(y.clone()))

        x_att = flash_attn_func(
            q=q.half(),  # flash requer float16
            k=k.half(),
            v=v.half(),
            dropout_p=self.dropout if self.training else 0.0,
            softmax_scale=1 / np.sqrt(self.head_dim),
            causal=self.causal,
            return_attn_probs=False,
            window_size=(self.window_size, self.window_size),
        )
        # para inferencia
        if not self.training or q.dtype != x_att.dtype:
            # se o modelo esta em float32 precisa voltar para float32
            for parametro in self.out_projection.parameters():
                x_att_ajustado = x_att.type(parametro.dtype)
                break
        else:
            x_att_ajustado = x_att

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att_reshaped = self.reshape_from_attention(x_att_ajustado)

        # projecao final
        x_att_projected = self.out_projection(self.dropout_projection(x_att_reshaped))

        return x_att_projected


# cross attention utilizado no stable diffusion 3
# https://stabilityai-public-packages.s3.us-west-2.amazonaws.com/Stable+Diffusion+3+Paper.pdf
class StableCrossFlashAttention(nn.Module):
    def __init__(
        self,
        num_heads: int = 8,
        embed_dim: int = 512,
        dropout=0.1,
        causal: bool = False,
        window_size: int = -1,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.causal = causal
        self.window_size = window_size

        assert (
            embed_dim % num_heads == 0
        ), "The number of dimensions must be divible by the number of heads"

        self.head_dim = embed_dim // num_heads
        self.out_projection = nn.Linear(2 * self.embed_dim, 2 * self.embed_dim)

        self.proj_q1 = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k1 = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v1 = nn.Linear(self.embed_dim, self.embed_dim)

        self.proj_q2 = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_k2 = nn.Linear(self.embed_dim, self.embed_dim)
        self.proj_v2 = nn.Linear(self.embed_dim, self.embed_dim)

        self.dropout_attention = nn.Dropout(dropout)
        self.dropout_projection = nn.Dropout(dropout)

        self.q1_norm = RMSNorm(dim=self.embed_dim)
        self.q2_norm = RMSNorm(dim=self.embed_dim)
        self.k1_norm = RMSNorm(dim=self.embed_dim)
        self.k2_norm = RMSNorm(dim=self.embed_dim)

    def reshape_for_attention(self, x):
        B, L, E = x.shape
        # shape x = [batch, len, embed_dim]
        x_view = x.contiguous().view((B, L, self.num_heads, self.head_dim))
        # virou [batch, len, heads, head_dim]
        return x_view

    def reshape_from_attention(self, x):
        B, L, H, HD = x.shape
        # virou [batch, len, heads, head_dim]
        x_view = x.contiguous().view(B, L, self.embed_dim)
        # virou [batch, len, embed_dim]
        return x_view

    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ):
        qx = self.reshape_for_attention(self.proj_q1(x.clone()))
        kx = self.reshape_for_attention(self.proj_k1(x.clone()))
        vx = self.reshape_for_attention(self.proj_v1(x.clone()))

        qy = self.reshape_for_attention(self.proj_q2(y.clone()))
        ky = self.reshape_for_attention(self.proj_k2(y.clone()))
        vy = self.reshape_for_attention(self.proj_v2(y.clone()))

        # ele normaliza as queries e keys usando rmsnorm
        qx = self.q1_norm(qx)
        qy = self.q2_norm(qy)
        kx = self.k1_norm(kx)
        ky = self.k2_norm(ky)

        # concatena todos os q, k, vs
        # coloca primeiro o cross depois o principal para funcionar com mascara causal (se quiser)
        q = torch.concat([qy, qx], dim=1)
        k = torch.concat([ky, kx], dim=1)
        v = torch.concat([vy, vx], dim=1)

        x_att = flash_attn_func(
            q=q.half(),  # flash requer float16
            k=k.half(),
            v=v.half(),
            dropout_p=self.dropout if self.training else 0.0,
            softmax_scale=1 / np.sqrt(self.head_dim),
            causal=self.causal,
            return_attn_probs=False,
            window_size=(self.window_size, self.window_size),
        )
        # para inferencia
        if not self.training or q.dtype != x_att.dtype:
            # se o modelo esta em float32 precisa voltar para float32
            for parametro in self.out_projection.parameters():
                x_att_ajustado = x_att.type(parametro.dtype)
                break
        else:
            x_att_ajustado = x_att

        # faz a concatenacao, volta para o shape [batch, len, embed_dim]
        x_att_reshaped = self.reshape_from_attention(x_att_ajustado)

        # projecao final
        x_att_projected = self.out_projection(self.dropout_projection(x_att_reshaped))

        return x_att_projected


class RotaryFlashDecoderBlock(nn.Module):
    def __init__(self, embed_dim, num_heads, hidden_size, rotary, dropout: int = 0.1):
        super().__init__()

        self.attention = RotaryMultiHeadFlashAttention(
            rotary=rotary,
            num_heads=num_heads,
            embed_dim=embed_dim,
            dropout=dropout,
            causal=True,
        )
        self.feedforward = FeedFowardBlock(
            embed_dim=embed_dim, hidden_size=hidden_size, dropout=dropout
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=embed_dim)
        self.norm_2 = nn.LayerNorm(normalized_shape=embed_dim)
        self.drop_skip_1 = nn.Dropout(dropout)
        self.drop_skip_2 = nn.Dropout(dropout)

    def forward(self, x):
        x = self.drop_skip_1(x) + self.attention(self.norm_1(x))
        x = self.drop_skip_2(x) + self.feedforward(self.norm_2(x))
        return x



import torch
from typing import Tuple


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.ndim
    assert 1 < ndim
    assert freqs_cis.shape == (x.shape[1], x.shape[-1])
    # insere dimensoes no meio mantedo a primeira nao batch e a ultima
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    return freqs_cis.view(*shape)


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    t = torch.arange(end, device=freqs.device, dtype=torch.float32)
    freqs = torch.outer(t, freqs)
    freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return freqs_cis


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    xq.shape = [B, L, H, D]
    """
    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
    freqs_cis = reshape_for_broadcast(freqs_cis, xq_)
    xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(3)
    xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)


def apply_cross_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Para ser usada quando q e k tem tamanhos diferentes
    xq.shape = [B, L, H, D]
    """
    q_len = xq_.shape[-3]
    k_len = xq_.shape[-3]

    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))

    freqs_cis_q = reshape_for_broadcast(freqs_cis[:q_len, :], xq_)
    freqs_cis_k = reshape_for_broadcast(freqs_cis[:k_len, :], xk_)

    xq_out = torch.view_as_real(xq_ * freqs_cis_q).flatten(3)
    xk_out = torch.view_as_real(xk_ * freqs_cis_k).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)



from typing import Any, Dict
from lightning.pytorch.utilities.types import OptimizerLRScheduler

import torch
import torch.nn as nn
import lightning.pytorch as pl
import bitsandbytes as bnb

from transformer.RetroMae.model import RetroMaeModel
from transformer.schedulers import CosineWarmupScheduler


class RetroMaeLightning(pl.LightningModule):
    def __init__(
        self,
        encoder_params: Dict,
        decoder_params: Dict,
        learning_rate: float = 1e-4,
        min_lr_percent: float = 0.1,
        warmup_steps: int = 500,
        total_training_steps: int = 1e6,
    ):
        super().__init__()

        self.save_hyperparameters()
        self.learning_rate = learning_rate
        self.min_lr_percent = min_lr_percent
        self.warmup_steps = warmup_steps
        self.total_training_steps = total_training_steps
        self.decoder_vocab_size = decoder_params["vocab_size"]

        self.model = RetroMaeModel(
            encoder_args=encoder_params, decoder_args=decoder_params
        )

    def training_step(self, batch, batch_idx):
        x, y, masked_mask = batch
        # flat the labels
        labels = y[masked_mask]
        # runs through the model
        encoder_logits, decoder_logits = self.model(x, masked_mask, y)
        mlm_loss = torch.nn.functional.cross_entropy(encoder_logits, labels)

        # para o decoder nao podemos considerar nem o primeiro indice, nem os tokens de padding
        y_no_cls = y[:, 1:]
        y_flat = y_no_cls.reshape((-1,))
        decoder_mask = torch.flatten(
            y_flat.nonzero()
        )  # retira o padding e o cls na maldade
        labels_decoder = y.reshape((-1, 1))
        labels_decoder = labels_decoder[decoder_mask, :].reshape((-1,))
        decoder_logits = decoder_logits.reshape((-1, self.decoder_vocab_size))
        decoder_logits = decoder_logits[decoder_mask, :]
        # pula o primeiro indice pq eh o cls token
        decoder_loss = torch.nn.functional.cross_entropy(decoder_logits, labels_decoder)

        loss = decoder_loss + mlm_loss
        self.log("train/decoder_loss", decoder_loss)
        self.log("train/mlm_loss", mlm_loss)
        self.log("train/loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y, masked_mask = batch
        # flat the labels
        labels = y[masked_mask]
        # runs through the model
        encoder_logits, decoder_logits = self.model(x, masked_mask, y)
        mlm_loss = torch.nn.functional.cross_entropy(encoder_logits, labels)

        # para o decoder nao podemos considerar nem o primeiro indice, nem os tokens de padding
        y_no_cls = y[:, 1:]
        y_flat = y_no_cls.reshape((-1,))
        decoder_mask = torch.flatten(
            y_flat.nonzero()
        )  # retira o padding e o cls na maldade
        labels_decoder = y.reshape((-1, 1))
        labels_decoder = labels_decoder[decoder_mask, :].reshape((-1,))
        decoder_logits = decoder_logits.reshape((-1, self.decoder_vocab_size))
        decoder_logits = decoder_logits[decoder_mask, :]
        # pula o primeiro indice pq eh o cls token
        decoder_loss = torch.nn.functional.cross_entropy(decoder_logits, labels_decoder)

        loss = decoder_loss + mlm_loss
        self.log("validation/decoder_loss", decoder_loss)
        self.log("validation/mlm_loss", mlm_loss)
        self.log("validationloss", loss)
        return loss

    def configure_optimizers(self):
        # optimizer = torch.optim.AdamW(self.parameters(), lr=self.learning_rate)
        optimizer = bnb.optim.AdamW8bit(
            self.parameters(),
            lr=self.learning_rate,
            betas=(0.9, 0.95),
            weight_decay=1e-2,
        )
        scheduler = CosineWarmupScheduler(
            optimizer=optimizer,
            warmup=self.warmup_steps,
            max_iters=self.total_training_steps,
            min_percent=self.min_lr_percent,
        )

        return [optimizer], [
            {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            }
        ]

    def lr_scheduler_step(self, scheduler, metric):
        scheduler.step()

    def forward(self, x, mask):
        return self.model(x, mask)

    def embed(self, x):
        return self.model.embed(x)



from typing import Any, Dict

import torch
import torch.nn as nn
import lightning.pytorch as pl

from transformer.vanilla.encoder import EncoderMLM
from transformer.schedulers import CosineWarmupScheduler


class VanillaEncoderMLMLightning(pl.LightningModule):
    def __init__(
        self,
        encoder_params: Dict,
        learning_rate: float = 1e-4,
        min_lr_percent: float = 0.1,
        warmup_steps: int = 500,
        total_training_steps: int = 1e6,
    ):
        super().__init__()

        self.save_hyperparameters()
        self.learning_rate = learning_rate
        self.min_lr_percent = min_lr_percent
        self.warmup_steps = warmup_steps
        self.total_training_steps = total_training_steps

        self.model = EncoderMLM(**encoder_params)

    def training_step(self, batch, batch_idx):
        x, y, masked_mask = batch
        # first get the masked_ids to use later
        # flattens the masked id so its easier to deal with
        masked_ids = torch.flatten(masked_mask.reshape((-1,)).nonzero())
        # also flat the labels
        labels = y.reshape((-1,))[masked_ids]
        # runs through the model
        out = self.model(x, masked_mask)
        loss = torch.nn.functional.cross_entropy(out, labels)
        self.log("train_loss", loss)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.learning_rate)
        # optimizer = bnb.optim.AdamW8bit(
        #     self.parameters(),
        #     lr=self.learning_rate,
        #     betas=(0.9, 0.95),
        #     weight_decay=1e-2,
        # )
        scheduler = CosineWarmupScheduler(
            optimizer=optimizer,
            warmup=self.warmup_steps,
            max_iters=self.total_training_steps,
            min_percent=self.min_lr_percent,
        )

        return [optimizer], [
            {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            }
        ]

    def lr_scheduler_step(self, scheduler, metric):
        scheduler.step()

    def generate_text(self, **kwargs):
        return self.model.generate_text(**kwargs)

    def forward(self, x, mask):
        return self.model(x, mask)



import torch
import numpy as np


def make_causal_mask(x: torch.Tensor):
    B, seq_len = x.shape
    mask = torch.triu(torch.ones((seq_len, seq_len))) - torch.eye(seq_len)
    mask = mask.masked_fill(mask == 1, float("-inf"))
    mask = mask.reshape((1, 1, seq_len, seq_len)).expand(B, -1, -1, -1)
    mask = mask.to(x.device)

    return mask


def _make_causal_mask(
    input_ids_shape: torch.Size,
    dtype: torch.dtype,
    device: torch.device,
    past_key_values_length: int = 0,
):
    """
    Make causal mask used for bi-directional self-attention.
    """
    bsz, tgt_len = input_ids_shape
    mask = torch.full((tgt_len, tgt_len), torch.finfo(dtype).min, device=device)
    mask_cond = torch.arange(mask.size(-1), device=device)
    mask.masked_fill_(mask_cond < (mask_cond + 1).view(mask.size(-1), 1), 0)
    mask = mask.to(dtype)

    if past_key_values_length > 0:
        mask = torch.cat(
            [
                torch.zeros(
                    tgt_len, past_key_values_length, dtype=dtype, device=device
                ),
                mask,
            ],
            dim=-1,
        )
    return mask[None, None, :, :].expand(
        bsz, 1, tgt_len, tgt_len + past_key_values_length
    )


def qkv_attention(q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, mask=None):
    """
    Recebe tensores no shape [B, H, L, D] para calculo da atencao
    """
    b, heads, len_tokens, embed_dim = q.shape
    k_t = torch.transpose(k, -1, -2)
    # shapes for q, k, v are [B, HEADS, SEQ_, HEAD_DIM]
    # for K_t we have [B, HEADS, HEAD_DIM, SEQ_K]
    qk = torch.einsum("bhsd, bhde -> bhse", q, k_t)
    # qk = torch.bmm(q, k_t)
    # shape of qk is [B, SEQ_Q, SEQ_K]
    if mask is not None:
        qk = qk + mask
    attention = torch.softmax(qk / np.sqrt(embed_dim), dim=-1)
    # [batch, heads, decoder_len, head_dim] * [batch, heasd, encoder_len, head_dim]
    full_attention = torch.einsum("bhde, bher -> bhdr", attention, v)
    return full_attention


def block_wise_parallel_attention(
    query, key, value, attention_mask, query_chunk_size, key_chunk_size
):
    """
    query, key, value - [batch, len, heads, head dim]
    """

    q = query / np.sqrt(query.shape[-1])

    batch, q_len, num_heads, dim_per_head = query.shape
    batch, kv_len, num_heads, dim_per_head = key.shape
    batch, kv_len, num_heads, dim_per_head = value.shape

    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size

    q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head)).half()
    k = key.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head)).half()
    v = value.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head)).half()

    q = q.transpose(1, 0)  # mudando o batch de lugar e colocando o indice antes
    k = k.transpose(1, 0)
    v = v.transpose(1, 0)

    outputs = []

    for query_idx in range(q.shape[0]):
        query_chunk = q[query_idx]

        prev_max_score = torch.tensor(float("-Inf")) * torch.ones(
            batch, query_chunk_size, num_heads, 1
        ).to(q.device).to(q.dtype)
        numerator = (
            torch.zeros((batch, query_chunk_size, num_heads, dim_per_head))
            .to(q.device)
            .to(q.dtype)
        )
        denominator = (
            torch.zeros((batch, query_chunk_size, num_heads, dim_per_head))
            .to(q.device)
            .to(q.dtype)
        )

        for kv_idx in range(k.shape[0]):
            key_chunk = k[kv_idx]
            value_chunk = v[kv_idx]

            attn_weights = torch.einsum("bqhd,bkhd->bqhk", query_chunk, key_chunk)
            # print(attn_weights.shape)
            mask_chunk = attention_mask[
                :,
                :,
                query_idx * query_chunk_size : (query_idx + 1) * query_chunk_size,
                kv_idx * key_chunk_size : (kv_idx + 1) * key_chunk_size,
            ]
            mask_chunk = mask_chunk.transpose(1, 2)

            # precisa checar se a mascara toda nao eh invalida para este bloco
            if mask_chunk.max() < 0:
                continue
            # print(mask_chunk.shape)

            attn_weights = attn_weights + mask_chunk

            max_score = torch.max(attn_weights, dim=-1, keepdim=True)[0]
            max_score = torch.maximum(prev_max_score, max_score).detach()

            exp_weights = torch.exp(attn_weights - max_score)
            exp_values = torch.einsum("bqhk,bkhd->bqhd", exp_weights, value_chunk)

            correction = torch.exp(prev_max_score - max_score)
            numerator = numerator * correction + exp_values
            denominator = denominator * correction + exp_weights.sum(
                dim=-1, keepdim=True
            )

            # nao esquece dessa merda
            prev_max_score = max_score

        outputs.append(numerator / denominator)

    outputs = torch.stack(outputs, dim=0)
    outputs = outputs.transpose(0, 1).reshape(batch, q_len, num_heads, dim_per_head)
    return outputs


def cross_batch_attention(
    query,
    key,
    value,
    cross_batch_range,
    max_query_att_lenght=1024,
    max_key_att_length=1024,
):
    batch_size, seq_len, num_heads, _ = query.shape

    num_attention = 1 + cross_batch_range

    cross_batch_rel_ids = torch.arange(0, -num_attention, -1).reshape(1, -1)
    batch_ids = torch.arange(0, batch_size).reshape(-1, 1)
    cross_batch_selector = cross_batch_rel_ids + batch_ids

    # other contexts
    cb_outer_keys = key[cross_batch_selector[:, 1:]]
    # ja abre as keys em comprimento, juntando os cross_batches
    cb_outer_keys = cb_outer_keys.reshape(
        (
            cb_outer_keys.shape[0],
            cb_outer_keys.shape[1] * cb_outer_keys.shape[2],
            cb_outer_keys.shape[3],
            cb_outer_keys.shape[4],
        )
    )
    # local + outer context
    cb_keys = torch.cat([cb_outer_keys, key], dim=1)

    cb_outer_values = value[cross_batch_selector[:, 1:]]
    cb_outer_values = cb_outer_values.reshape(
        (
            cb_outer_values.shape[0],
            cb_outer_values.shape[1] * cb_outer_values.shape[2],
            cb_outer_values.shape[3],
            cb_outer_values.shape[4],
        )
    )
    cb_values = torch.cat([cb_outer_values, value], dim=1)

    # precisa de uma mascara para atencao
    mask = _make_causal_mask(
        input_ids_shape=query.shape[:2],
        dtype=query.dtype,
        device=query.device,
        past_key_values_length=cb_outer_keys.shape[1],
    )

    mask = mask.repeat(1, num_heads, 1, 1)

    attn = block_wise_parallel_attention(
        query=query,
        key=cb_keys,
        value=cb_values,
        attention_mask=mask,
        query_chunk_size=max_query_att_lenght,
        key_chunk_size=max_key_att_length,
    )

    return attn



from multiprocessing import Process
from typing import List
import time

import numpy as np


def run_pool_of_processes(
    estoque_processos: List[Process],
    num_workers: int,
    timeout_delaying_process: bool = False,
    process_timeout: int = 60,
):
    quantidade_estoque = 0
    quantidade_ativo = 0
    processos_ativos = []
    quantidade_estoque_original = len(estoque_processos)
    tempo_ultimo = time.time()

    while len(estoque_processos) > 0 or len(processos_ativos) > 0:
        matou_algum = True
        while matou_algum:  # itera para ir matando os caras que finalizaram
            matou_algum = False
            for i in range(len(processos_ativos)):
                if not processos_ativos[i].is_alive():
                    p = processos_ativos.pop(i)
                    p.terminate()
                    del p
                    tempo_ultimo = time.time()
                    break

        while len(processos_ativos) < num_workers and len(estoque_processos) > 0:
            indice = np.random.randint(len(estoque_processos))
            p = estoque_processos.pop(indice)
            p.start()
            processos_ativos.append(p)

        # isso aqui eh so para printar e ir acompanhando o quanto ja rodou
        if quantidade_estoque != len(estoque_processos) or quantidade_ativo != len(
            processos_ativos
        ):
            quantidade_estoque = len(estoque_processos)
            quantidade_ativo = len(processos_ativos)
            percentual_completo = (
                1
                - (quantidade_estoque + quantidade_ativo) / quantidade_estoque_original
            )
            percentual_completo *= 100
            print(
                f"Estoque: {len(estoque_processos)}\tAtivos: {len(processos_ativos)}\tPercent: {percentual_completo:3.4f}"
            )

            # quando a parada demora muito para processar vamos dar timeout
            if timeout_delaying_process:
                if time.time() - tempo_ultimo >= process_timeout:
                    if len(processos_ativos) > 0:
                        p = processos_ativos.pop(0)
                        p.terminate()
                        del p
                        tempo_ultimo = time.time()

        time.sleep(0.1)



import torch


def count_parameters(model):
    n_params = sum(p.numel() for p in model.parameters())
    print("number of parameters: %.2fM" % (n_params / 1e6,))


def create_causal_mask(query_len, key_len):
    if key_len > query_len:
        triangular_shape = (query_len, query_len)
        excedente = key_len - query_len
        mask = torch.concat(
            [
                torch.ones(query_len, excedente),
                torch.tril(torch.ones(triangular_shape)),
            ],
            dim=1,
        )
    elif query_len > key_len:
        raise NotImplementedError("O tamanho da query esta maior que da key")
    else:
        triangular_shape = (query_len, key_len)
        mask = torch.tril(torch.ones(triangular_shape))

    return mask.reshape(1, 1, query_len, key_len)


class TransformerCache:
    def __init__(self, cache_max_len: int = 8192):
        self.max_len = cache_max_len
        self.k = None
        self.v = None

    def update_v_cache(self, x):
        """
        x - [batch, len, heads, head_dim]
        """
        if self.v is None:
            self.v = x
        else:
            self.v = torch.cat([self.v, x], dim=2)

        if self.v.shape[2] > self.max_len:
            self.v = self.v[:, :, -self.max_len :, :]

    def update_k_cache(self, x):
        """
        x - [batch, len, heads, head_dim]
        """
        if self.k is None:
            self.k = x
        else:
            self.k = torch.cat([self.k, x], dim=2)

        if self.k.shape[2] > self.max_len:
            self.k = self.k[:, :, -self.max_len :, :]

    def add_v_cache(self, x, update_cache: bool = True):
        """
        x - [batch, heads, len, head_dim]
        """

        if self.v is not None:
            ans = torch.cat([self.v, x], dim=2)
        else:
            ans = x

        if update_cache:
            self.update_v_cache(x)
        return ans

    def add_k_cache(self, x, update_cache: bool = True):
        """
        x - [batch, heads, len, head_dim]
        """
        if self.k is not None:
            ans = torch.cat([self.k, x], dim=2)
        else:
            ans = x

        if update_cache:
            self.update_k_cache(x)
        return ans



import math

import torch
import torch.nn as nn


class NewGELU(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, x):
        return (
            0.5
            * x
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))
                )
            )
        )



import torch
import torch.nn as nn

from transformer.transformer_blocks import EncoderBlock


class PositionalEncoding(nn.Module):
    def __init__(
        self, model_dimension, dropout_probability, expected_max_sequence_length=5000
    ):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout_probability)

        position_id = torch.arange(0, expected_max_sequence_length).unsqueeze(1)
        frequencies = torch.pow(
            10000.0,
            -torch.arange(0, model_dimension, 2, dtype=torch.float) / model_dimension,
        )

        positional_encodings_table = torch.zeros(
            expected_max_sequence_length, model_dimension
        )
        positional_encodings_table[:, 0::2] = torch.sin(
            position_id * frequencies
        )  # sine on even positions
        positional_encodings_table[:, 1::2] = torch.cos(
            position_id * frequencies
        )  # cosine on odd positions

        self.register_buffer("positional_encodings_table", positional_encodings_table)

    def forward(self, embeddings_batch):
        assert (
            embeddings_batch.ndim == 3
            and embeddings_batch.shape[-1] == self.positional_encodings_table.shape[1]
        ), f"Expected (batch size, max token sequence length, model dimension) got {embeddings_batch.shape}"

        positional_encodings = self.positional_encodings_table[
            : embeddings_batch.shape[1]
        ]

        return self.dropout(embeddings_batch + positional_encodings)


class Encoder(nn.Module):
    def __init__(
        self,
        vocab_size: int,
        num_layers: int,
        embed_dim: int,
        hidden_size: int,
        num_heads: int,
        dropout: float,
        max_context_size: int = 1024,
    ):
        super().__init__()

        self.pos_encoding = PositionalEncoding(
            model_dimension=embed_dim,
            dropout_probability=dropout,
            expected_max_sequence_length=max_context_size,
        )
        self.embedding = nn.Embedding(
            num_embeddings=vocab_size, embedding_dim=embed_dim
        )

        self.layers = nn.ModuleList()
        for _ in range(num_layers):
            self.layers.append(
                EncoderBlock(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    hidden_size=hidden_size,
                    dropout=dropout,
                )
            )

    def forward(self, x):
        x = self.embedding(x)
        x = self.pos_encoding(x)

        for layer in self.layers:
            x = layer(x)

        return x


class EncoderMLM(nn.Module):
    def __init__(
        self,
        vocab_size: int,
        num_layers: int,
        embed_dim: int,
        hidden_size: int,
        num_heads: int,
        dropout: float,
        max_context_size: int = 1024,
    ):
        super().__init__()

        self.embed_dim = embed_dim

        self.encoder = Encoder(
            vocab_size=vocab_size,
            num_layers=num_layers,
            num_heads=num_heads,
            embed_dim=embed_dim,
            hidden_size=hidden_size,
            dropout=dropout,
            max_context_size=max_context_size,
        )

        self.mlm_head = nn.Linear(embed_dim, vocab_size)

    def forward(self, x, mask):
        # first get the masked_ids to use later
        # flattens the masked id so its easier to deal with
        masked_ids = torch.flatten(mask.reshape((-1,)).nonzero())
        # get all hidden states
        last_hidden_states = self.encoder(x)
        # flatten everything so we can compute everything at once
        all_hidden_states = last_hidden_states.reshape(-1, self.embed_dim)
        # get only the masked hidden states
        masked_hidden_states = all_hidden_states[masked_ids, :]
        # predicts only the masked tokens
        logits = self.mlm_head(masked_hidden_states)

        return logits



from setuptools import find_packages, setup

# OBS: "flash-attn==2.5.8" precisa ser instalado depois pq ele requer
# que alguns pacotes estejam instalados antes como wheel, ninja e packaging
# instala ele depois com pip install flash-attn==2.5.8 --no-build-isolation

_deps = [
    "numpy==2.0.0",
    "charylu-tokenizer==0.0.5",
    "tqdm",
    "torch==2.3.1",
    "lightning==2.3.0",
    "wheel",
    "bitsandbytes==0.43.1",
    "packaging==24.1",
    "ninja==1.11.1.1",
    "torchmetrics==1.4.0.post0",
    "deepspeed==0.14.4",
]

setup(
    name="charylu-models",
    packages=find_packages(include=["charylumodels"]),
    # include_package_data=True,
    package_data={"": ["transformer/*.py", "models/**/*.py"]},
    version="0.0.4",
    description="Biblioteca de modelos implemantados por Luis Chary",
    author="Luis Felipe Chary",
    install_requires=_deps,
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
)



from typing import List
from pathlib import Path
import pickle
from multiprocessing import Process
import time

from tqdm import tqdm
import pandas as pd
import numpy as np

from charylutokenizer.charylutokenizer import CharyluTokenizer


def tokenize_partition(
    partition_ref: str, partition_num: int, nome_base: str, tokenizer: CharyluTokenizer
):
    root_path = Path("/media/luis/BIGGER/datasets/nlp_datasets/export")
    metadata_folder = root_path / "tokenized_metadata"
    if not metadata_folder.exists():
        metadata_folder.mkdir(parents=True, exist_ok=True)
    metadata_path = metadata_folder / f"{nome_base}_{partition_num}.pq"
    if metadata_path.exists():
        return

    binary_folder = root_path / "tokenized_512" / nome_base
    if not binary_folder.exists():
        binary_folder.mkdir(parents=True, exist_ok=True)

    binary_path = binary_folder / f"{nome_base}_{partition_num}.npz"
    if binary_path.exists():
        return

    texts = pd.read_parquet(partition_ref)
    chunks = []

    tokenized = []
    tokens_count = 0
    for _, row in texts.iterrows():
        texto = row.text
        # tokeniza adicionando so o fim, o comeco vai vir depois
        tokenized += tokenizer.tokenize(texto) + [2]
        while len(tokenized) > 511:
            chunk = np.array(
                [1] + tokenized[:511]
            )  # adiciona o comeco aqui para o retroMAE
            chunks.append(chunk)
            tokens_count += 512
            tokenized = tokenized[511:]

    # no final guarda o que sobrou, adicionando padding ja
    tokenized = [1] + tokenized
    diff = 512 - len(tokenized)
    tokenized += [0 for _ in range(diff)]
    chunks.append(np.array(tokenized))
    tokens_count += 512

    # empilha os chunks
    chunks = np.vstack(chunks, dtype=int)
    # salva bonitinho
    np.savez_compressed(binary_path.as_posix(), chunks)
    # adiciona a referencia nos metadados
    metadados = {
        "partition_ref": [partition_ref],
        "tokenized_512_ref": [binary_path.as_posix()],
        "num_tokens": [tokens_count],
    }
    metadados = pd.DataFrame(metadados)
    metadados.to_parquet(metadata_path.as_posix())


if __name__ == "__main__":
    # tokenizer = CharyluTokenizer(
    #     tokenizer_path="/home/luis/projetos/luis_transformers/artifacts/charylu_nocode/tokenizer_2024_90k.json"
    # )

    # df = pd.read_parquet(
    #     "/media/luis/BIGGER/datasets/nlp_datasets/metadata/all_metadata_nodup_lang.pq"
    # )

    # df = df[df.duplicado_80 == 0].reset_index(drop=True)
    # df = df[df.lang.isin(["pt", "en", "fr", "it", "es"])].reset_index(drop=True)
    # df = df[df.tipo != "code"].reset_index(drop=True)
    # partition_num = 0
    # processos = []
    # ja_foi = set()
    # for _, row in tqdm(df.iterrows(), total=len(df)):
    #     partition = row.partition_ref
    #     nome_base = row.base
    #     if partition not in ja_foi:
    #         processos.append(
    #             Process(
    #                 target=tokenize_partition,
    #                 args=(partition, partition_num, nome_base, tokenizer),
    #             )
    #         )
    #         partition_num += 1
    #         ja_foi.add(partition)

    # df = None
    # del df

    # run_pool_of_processes(
    #     estoque_processos=processos,
    #     num_workers=32,
    #     timeout_delaying_process=True,
    #     process_timeout=60 * 30,
    # )

    # agora junta todos os metadados
    todos = []
    for p in Path(
        "/media/luis/BIGGER/datasets/nlp_datasets/export/tokenized_metadata"
    ).glob("*.pq"):
        todos.append(pd.read_parquet(p.as_posix()))
    todos = pd.concat(todos).reset_index(drop=True)

    def nome_base(p):
        p = Path(p)
        return p.parent.stem

    todos["base"] = todos.partition_ref.apply(lambda x: nome_base(x))
    todos.to_parquet(
        "/media/luis/BIGGER/datasets/nlp_datasets/export/tokenized_metadata.pq"
    )



from typing import List, Dict
from pathlib import Path
import pickle
import math


import torch
from torch.utils.data import Dataset
import numpy as np
import pandas as pd


class TokenizedEncoderMLMDataset(Dataset):
    def __init__(
        self,
        metadata_path: str,
        max_len: int = 512,
        mask_prob: float = 0.15,
        limit: int = None,
        mask_token_id: int = 1,
        pad_token_id: int = 0,
        special_tokens: List[int] = [],
        vocab_size: int = 60000,
    ):
        super().__init__()

        self.max_len = max_len
        self.mask_prob = mask_prob
        self.mask_token_id = mask_token_id
        self.pad_token_id = pad_token_id
        self.special_tokens = special_tokens
        self.vocab_size = vocab_size

        self.metadata = pd.read_parquet(metadata_path)

        if limit is not None:
            self.metadata = self.metadata[self.metadata.indice <= limit]

    def __len__(self):
        return self.metadata.indice.max()

    def __getitem__(self, index):
        # primeiro localiza qual particao tokenizada
        linha_metadata = self.metadata[self.metadata.indice > index].reset_index(
            drop=True
        )
        matriz = np.load(linha_metadata.loc[0, "tokenized_512_ref"])["arr_0"]
        # qual o indice do cara na matriz?
        final_matriz = linha_metadata.loc[0, "indice"]
        tras_para_frente = final_matriz - index
        tokenized = matriz[-tras_para_frente]
        # agora so seguir

        # before masking its nice to have everything in the right size
        if len(tokenized) < self.max_len:
            diff = self.max_len - len(tokenized)
            tokenized += [self.pad_token_id for _ in range(diff)]
        elif len(tokenized) > self.max_len:
            tokenized = tokenized[: self.max_len]
        # transform into tensor
        tokenized = torch.from_numpy(tokenized)
        # mask time
        probas = torch.rand(tokenized.shape)
        mask = (probas < self.mask_prob) * (tokenized != self.pad_token_id)
        # special tokens
        for special_token in self.special_tokens:
            mask = mask * (tokenized != special_token)

        # now mask tokenized with the mask we just created
        masked = torch.clone(tokenized).type(torch.int)
        masked_ids = torch.flatten(mask.nonzero())
        masked_ids_list = masked_ids.tolist()
        # 80% will be replaced by the mask token
        # 10% no change
        # 10% replaced by random token
        original_masked_tokens = tokenized[masked_ids_list]
        replace_masked_tokens = self.generate_mlm_tokens(
            original_masked_tokens.tolist()
        )
        masked[masked_ids_list] = replace_masked_tokens

        return masked, tokenized, mask

    def generate_mlm_tokens(self, original_tokens: List[int]):
        # 80% will be replaced by the mask token
        # 10% no change
        # 10% replaced by random token
        replace_tokens = np.random.rand(len(original_tokens))
        for i in range(len(original_tokens)):
            if replace_tokens[i] <= 0.8:
                replace_tokens[i] = self.mask_token_id
            elif replace_tokens[i] <= 0.9:
                replace_tokens[i] = np.random.randint(self.vocab_size)
            else:
                replace_tokens[i] = original_tokens[i]
        return torch.from_numpy(replace_tokens).type(torch.int)



import torch
import numpy as np


class CosineWarmupScheduler(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, warmup, max_iters, min_percent=0.01):
        self.warmup = warmup
        self.max_num_iters = max_iters
        self.min_percent = min_percent
        super().__init__(optimizer)

    def get_lr(self):
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_num_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch * 1.0 / self.warmup
        else:
            lr_factor = max(lr_factor, self.min_percent)
        return lr_factor


class LinearWarmupScheduler(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, warmup, max_iters, min_percent=0.01):
        self.warmup = warmup
        self.max_num_iters = max_iters
        self.min_percent = min_percent
        super().__init__(optimizer)

    def get_lr(self):
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        if epoch <= self.warmup:
            lr_factor = max(epoch * 1.0 / self.warmup, self.min_percent)
        else:
            lr_factor = max(
                1 - ((epoch - self.warmup) * 1.0 / self.max_num_iters), self.min_percent
            )

        return lr_factor
